// PrestoChannelFactory.cs
//
// PRESTO Starter Kit for Microsoft .NET 3.0
// 2007 Microsoft France
// Published under the CECILL-B Free Software license agreement.
// (http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.txt)
// 
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, 
// WHETHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED 
// WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE. 
// THE ENTIRE RISK OF USE OR RESULTS IN CONNECTION WITH THE USE OF THIS CODE 
// AND INFORMATION REMAINS WITH THE USER. 
//

using System;
using System.Configuration;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Text;
using System.Xml;
using System.Xml.Serialization;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Configuration;
using System.ServiceModel.Description;

using System.ServiceModel.Security;
using System.Security.Cryptography.X509Certificates;
using System.Net.Security;

using dgme.finances.gouv.fr.presto;

using Microsoft.Presto.StarterKit.Constants;

namespace Microsoft.Presto.StarterKit
{
    public class PrestoChannelFactory
    {
        private PrestoSoapNodeConfig config;
        public PrestoSoapNodeConfig Config
        {
            get { return config; }
            set { config = value; }
        }
        
        // Send a messsage to a Service endpoint for the IPresto Interface (Reliable One-Way MEP) 
        public void SendMessageToIPrestoEndpoint(string payloadFilename)
        {
            try
            {
                allData message = CreateMessage(payloadFilename);
                SubmitMessageOneWay(message);

                OutputTrace.Log("Submit One-Way call succeeded w/errors");
            }
            catch (Exception ex)
            {
                LogException(ex);
            }
        }

        // Send a messsage to a Service endpoint for the IPresto2 Interface (Reliable Request-Reply MEP)
        public void SendMessageToIPresto2Endpoint(string payloadFilename, Boolean saveResponse)
        {
            try
            {
                allData request = CreateMessage(payloadFilename);
                allData reply = SubmitMessage(request);
                if (saveResponse)
                {
                    SaveMessage(reply);
                }

                OutputTrace.Log("Request-Reply call succeeded w/errors");
            }
            catch (Exception ex)
            {
                LogException(ex);
            }
        }

        private void SubmitMessageOneWay(allData message)
        {
            ChannelFactory<IPresto> factory = null;
            EndpointAddress endPointAddress = null;

            // If the service as to authentify itself, the remote endpoint provides DNS claim 'PRESTO Target Application'.
            // As the expected DNS identity of the remote endpoint is 'localhost' 
            // it is neccessary to explicitly specifying DNS identity 'PRESTO Target Application' as the Identity property of EndpointAddress
            // when creating channel proxy to avoid an error about Identity check failed for outgoing message

            if (!(config.NoAuthentication))
            {
                EndpointIdentity endpointIdentity = EndpointIdentity.CreateDnsIdentity(config.ServiceIdentity);
                AddressHeader[] aAH = new AddressHeader[] { };

                if (config.AutoCompleteUris)
                {
                    endPointAddress = new EndpointAddress(new Uri(config.ServiceEndpointUri + "/" + EndpointUri.oneWayRelativeAddress), endpointIdentity, aAH);
                }
                else
                {
                    endPointAddress = new EndpointAddress(new Uri(config.ServiceEndpointUri), endpointIdentity, aAH);
                }
            }
            else
            {
                if (config.AutoCompleteUris)
                {
                    endPointAddress = new EndpointAddress(new Uri(config.ServiceEndpointUri + "/" + EndpointUri.oneWayRelativeAddress));
                }
                else
                {
                    endPointAddress = new EndpointAddress(new Uri(config.ServiceEndpointUri));
                }
            }

            if (config.UseMetadataExchange)
            {
                Uri mexUri = new Uri(config.ServiceEndpointMexUri);
                ContractDescription contract = ContractDescription.GetContract(typeof(IPresto));

                EndpointAddress mexEndpointAddress = new EndpointAddress(mexUri);
                ServiceEndpointCollection endpoints = MetadataResolver.Resolve(contract.ContractType, mexEndpointAddress);

                foreach (ServiceEndpoint endpoint in endpoints)
                {
                    if (endpoint.Contract.Namespace.Equals(contract.Namespace) && endpoint.Contract.Name.Equals(contract.Name))
                    {
                        factory = new ChannelFactory<IPresto>(endpoint.Binding, endpoint.Address);
                        GetConfigSecurityFromMex(endpoint);

                        AddCertificatesToFactoryCredentialsIfRequired(factory);
                        break;
                    }
                }
            }
            else if (config.UseConfigurationFile)
            {
                config.AddressingVersion = PrestoAddressingVersion.WSAddressingAugust2004;
                config.UseDigitalSignature = false;
                config.UseEncryption = false;

                string endPointConfigurationName = config.GetEndpointNameFromConfigurationFile(true);
                
                factory = new ChannelFactory<IPresto>(endPointConfigurationName, endPointAddress);
            }
            else
            {
                factory = new ChannelFactory<IPresto>(config.GenerateBinding(), endPointAddress);
                AddCertificatesToFactoryCredentialsIfRequired(factory);
            }

            if (config.UseSoapIntermediary)
            {
                if (config.AutoCompleteUris)
                {
                    factory.Endpoint.Behaviors.Add(new ClientViaBehavior(new Uri(config.SoapIntermediaryEndpointUri + "/" + EndpointUri.oneWayRelativeAddress)));
                }
                else
                {
                    factory.Endpoint.Behaviors.Add(new ClientViaBehavior(new Uri(config.SoapIntermediaryEndpointUri)));
                }
            }

            if (config.UseReliableMessaging)
            {
                ContractDescription contractDescription = factory.Endpoint.Contract;
                contractDescription.SessionMode = SessionMode.Required;
                OperationDescription operationDescription = contractDescription.Operations[0];
                operationDescription.IsInitiating = true;
                operationDescription.IsTerminating = true;
                
            }

            IPresto prestoInterface = factory.CreateChannel();
            using (new OperationContextScope((IContextChannel)prestoInterface))
            {
                OperationContext.Current.OutgoingMessageHeaders.MessageId = new UniqueId();

                submit1WayMessage sr = new submit1WayMessage();
                sr.submitOneWay = new submit1Way();
                sr.submitOneWay.testDocIn = message;

                prestoInterface.submitOneWay(sr);

                //IContextChannel channel = (IContextChannel)prestoInterface;
                //channel.Close();
            }
        }

        private allData SubmitMessage(allData message)
        {
            ChannelFactory<IPresto2> factory = null;
            EndpointAddress endPointAddress = null;

            // Set the Service endpoint address with its Identity in case of XMLDSIG support (signing)
            if (config.UseDigitalSignature)
            {
                EndpointIdentity endpointIdentity = EndpointIdentity.CreateDnsIdentity(config.ServiceIdentity);
                AddressHeader[] aAH = new AddressHeader[] { };

                if (config.AutoCompleteUris)
                {
                    endPointAddress = new EndpointAddress(new Uri(config.ServiceEndpointUri + "/" + EndpointUri.requestReplyRelativeAddress), endpointIdentity, aAH);
                }
                else
                {
                    endPointAddress = new EndpointAddress(new Uri(config.ServiceEndpointUri), endpointIdentity, aAH);
                }
            }
            else
            {
                if (config.AutoCompleteUris)
                {
                    endPointAddress = new EndpointAddress(new Uri(config.ServiceEndpointUri + "/" + EndpointUri.requestReplyRelativeAddress));
                }
                else
                {
                    endPointAddress = new EndpointAddress(new Uri(config.ServiceEndpointUri));
                }
            }

            if (config.UseMetadataExchange)
            {
                Uri mexUri = new Uri(config.ServiceEndpointMexUri);
                ContractDescription contract = ContractDescription.GetContract(typeof(IPresto2));

                EndpointAddress mexEndpointAddress = new EndpointAddress(mexUri);
                ServiceEndpointCollection endpoints = MetadataResolver.Resolve(contract.ContractType, mexEndpointAddress);

                foreach (ServiceEndpoint endpoint in endpoints)
                {
                    if (endpoint.Contract.Namespace.Equals(contract.Namespace) && endpoint.Contract.Name.Equals(contract.Name))
                    {
                        factory = new ChannelFactory<IPresto2>(endpoint.Binding, endpoint.Address);

                        GetConfigSecurityFromMex(endpoint);
                        
                        AddCertificatesToFactoryCredentialsIfRequired(factory);

                        break;
                    }
                }
            }
            else if (config.UseConfigurationFile)
            {
                config.AddressingVersion = PrestoAddressingVersion.WSAddressingAugust2004;
                config.UseDigitalSignature = false;
                config.UseEncryption = false;
                config.NoAuthentication = true;
                string endPointConfigurationName = config.GetEndpointNameFromConfigurationFile(false);
 
                factory = new ChannelFactory<IPresto2>(endPointConfigurationName, endPointAddress);
            }
            else
            {
                factory = new ChannelFactory<IPresto2>(config.GenerateBinding(), endPointAddress);
                AddCertificatesToFactoryCredentialsIfRequired(factory);
            }

            if (config.UseSoapIntermediary)
            {
                if (config.AutoCompleteUris)
                {
                    factory.Endpoint.Behaviors.Add(new ClientViaBehavior(new Uri(config.SoapIntermediaryEndpointUri + "/" + EndpointUri.requestReplyRelativeAddress)));
                }
                else
                {
                    factory.Endpoint.Behaviors.Add(new ClientViaBehavior(new Uri(config.SoapIntermediaryEndpointUri)));
                }
            }

            if (config.UseReliableMessaging)
            {
                ContractDescription contractDescription = factory.Endpoint.Contract;
                contractDescription.SessionMode = SessionMode.Required;
                OperationDescription operationDescription = contractDescription.Operations[0];
                operationDescription.IsInitiating = true;
                operationDescription.IsTerminating = true;

            }
            
            IPresto2 prestoInterface = factory.CreateChannel();
            using (new OperationContextScope((IContextChannel)prestoInterface))
            {
                OperationContext.Current.OutgoingMessageHeaders.MessageId = new UniqueId();

                submitRequestMessage sr = new submitRequestMessage();
                sr.submit = new submit();
                sr.submit.testDocIn = message;

                submitResponseMessage sresponse = prestoInterface.submit(sr);

                // The messageId header only applies to the first message in the sequence. Subsequent messages
                // are sent to the ReplyTo address of the Client endpoint and with the RelatesTo header set to 
                // the original MessageId.
                // OperationContext.Current.IncomingMessageHeaders.RelatesTo

                if ((sresponse != null) && (sresponse.submitResponse != null))
                {
                    return sresponse.submitResponse.testDocOut;
                }

                //IContextChannel channel = (IContextChannel)prestoInterface;
                //channel.Close();

                return null;
            }
        }

        private void GetConfigSecurityFromMex(ServiceEndpoint endpoint)
        {
            // Get the security requirements for sending message and configure the factory accordingly
            ISecurityCapabilities isc = endpoint.Binding.GetProperty<ISecurityCapabilities>(new BindingParameterCollection());

            // Trouble is that AnonymousForCertificate seems to give the same results as MutualCertificate
            // when asking SupportsClientAuthentication and SupportsServerAuthentication
            if (isc.SupportsClientAuthentication)
            {
                config.MutualCertificate = true;
                config.AnonymousForCertificate = false;
                config.NoAuthentication = false;
            }
            else if (isc.SupportsServerAuthentication)
            {
                config.AnonymousForCertificate = true;
                config.MutualCertificate = false;
                config.NoAuthentication = false;
            }
            else
            {
                config.NoAuthentication = true;
                config.AnonymousForCertificate = false;
                config.MutualCertificate = false;
            }

            if (isc.SupportedRequestProtectionLevel == ProtectionLevel.Sign)
            {
                config.UseDigitalSignature = true;
            }
            if (isc.SupportedRequestProtectionLevel == ProtectionLevel.EncryptAndSign)
            {
                config.UseEncryption = true;
            }
        }

        // Create message based on the values filled in the UI
        private allData CreateMessage(string payloadFilename)
        {
            allData message = new allData();
            message.intStrElement = new allDataIntStrElement();

            if (File.Exists(payloadFilename))
            {
                FileStream fstream = null;
                try
                {
                    fstream = File.Open(payloadFilename, FileMode.Open);
                    message.binaryElement = new byte[fstream.Length];
                    fstream.Read(message.binaryElement, 0, (int)fstream.Length);
                }
                catch (Exception ex)
                {
                    throw new Exception(ex.Message);
                }
                finally
                {
                    if (fstream != null) fstream.Close();
                }

                message.intStrElement = new allDataIntStrElement();
                message.intStrElement.intElement = "" + message.binaryElement.Length;
                message.intStrElement.strElement = payloadFilename;

                FileInfo fileInfo = new FileInfo(payloadFilename);
                message.anyElement = fileInfo.Name;
            }
            else
            {
                throw new Exception("Filename [" + payloadFilename + "] doesn't exist");
            }

            return message;
        }

        private void SaveMessage(allData message)
        {
            if (message == null)
            {
                OutputTrace.Log("<null> reply message!?");
            }
            else if (message.binaryElement == null)
            {
                OutputTrace.Log("<null> reply message binary element!?");
            }
            else
            {
                string fileName = Guid.NewGuid() + ".txt";
                if (File.Exists(fileName)) File.Delete(fileName);
                FileInfo fileInfo = new FileInfo(fileName);

                FileStream stream = File.Create(fileName);
                stream.Write(message.binaryElement, 0, message.binaryElement.Length);
                stream.Close();

                OutputTrace.Log("reply message saved as " + fileInfo.FullName);
            }
        }

        private void AddCertificatesToFactoryCredentialsIfRequired(ChannelFactory factory)
        {
            if (!(config.NoAuthentication))
            {
                if (!(config.UseMetadataExchange))
                {
                    // Set the X.509 Service Certificate as Credential for the factory
                    factory.Credentials.ServiceCertificate.SetDefaultCertificate(StoreLocation.LocalMachine, StoreName.My,
                                X509FindType.FindByThumbprint, config.ServiceCertificateThumbprint);
                }

                // Setting the CertificateValidationMode to PeerTrust or PeerOrChainTrust means that if the  
                // certificate is in the Trusted People store, then it will be trusted without performing a validation 
                // of the certificate's issuer chain. Such a setting is used here for convenience so that the 
                // sample can be run without having to have certificates issued by a certificate authority (CA).
                // This setting is less secure than the default, ChainTrust. The security implications of this 
                // setting should be carefully considered before using PeerTrust or PeerOrChainTrust in production code. 
                factory.Credentials.ServiceCertificate.Authentication.CertificateValidationMode = X509CertificateValidationMode.ChainTrust;

                // The PRESTO Starter Kit X.509 Certificates have been generated with the Certificate Creation tool
                // (makecert.exe). These certificates are for development/testing purposes only and do NOT have a CDP
                // extension. As a result, the revocation test has to been turned off, otherwise, such certificates are
                // considered expired. In production code, the following line MUST be commented.
                factory.Credentials.ServiceCertificate.Authentication.RevocationMode = X509RevocationMode.NoCheck;

                if (config.MutualCertificate)
                {
                    // Set the X.509 Client Signing Certificate as Credential for the factory
                    factory.Credentials.ClientCertificate.SetCertificate(StoreLocation.LocalMachine, StoreName.My,
                                X509FindType.FindByThumbprint, config.ClientCertificateThumbprint);

                    // Adding the following behavior should have the same effect than the previous API call
                    //// Extract the ClientCredentials behaviour or create one
                    //ClientCredentials clientCredentials = factory.Endpoint.Behaviors.Find<ClientCredentials>();
                    //if (clientCredentials == null)
                    //{
                    //    clientCredentials = new ClientCredentials();
                    //    factory.Endpoint.Behaviors.Add(clientCredentials);
                    //}

                    ////// Set the Service X.509 certificate
                    //clientCredentials.ClientCertificate.SetCertificate(StoreLocation.LocalMachine, StoreName.My,
                    //            X509FindType.FindByThumbprint, config.ClientCertificateThumbprint);

                }
            }
 
            if (config.UseEncryption)
            {
                factory.Endpoint.Contract.ProtectionLevel = ProtectionLevel.EncryptAndSign;
            }
            else
            {
                if (config.UseDigitalSignature)
                {
                    factory.Endpoint.Contract.ProtectionLevel = ProtectionLevel.Sign;
                }
                else
                {
                    factory.Endpoint.Contract.ProtectionLevel = ProtectionLevel.None;
                }
            }
        }

        private void LogException(Exception ex)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("An error has been encountered: \n" 
                       + "Type : " + ex.GetType().ToString() + "\n" 
                       + ex.Message + "\n" + ex.StackTrace);
            OutputTrace.Log(sb.ToString());
            
            if (ex.InnerException != null)
            {
                StringBuilder sbInner = new StringBuilder();
                sbInner.Append("An error has been encountered: \n"
                              + "Type : " + ex.InnerException.GetType().ToString() + "\n"
                              + ex.InnerException.Message + "\n" + ex.InnerException.StackTrace);
                OutputTrace.Log(sbInner.ToString());
            }
        }
    }
}